/* $XConsortium: SmeCMP.c,v 1.19 94/04/17 20:12:49 kaleb Exp $ */

/*
   Copyright (c) 1989, 1994  X Consortium

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
   AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the X Consortium shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from the X Consortium.
 */

/*
 * SmeCMP.c - Source code file for CMP Menu Entry object.
 *
 * Date:    September 26, 1989
 *
 * By:      Chris D. Peterson
 *          MIT X Consortium 
 *          kit@expo.lcs.mit.edu
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xos.h>

/* needed for abs() */
#ifdef X_NOT_STDC_ENV
extern int abs();
#else
#include <stdlib.h>
#endif

#include <X11/Xmu/Drawing.h>

#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Cardinals.h>
#include "SmeCMPP.h"

#include <stdio.h>

#define ONE_HUNDRED 100

#define offset(field) XtOffsetOf(SmeCMPRec, sme_cmp.field)

static XtResource resources[] =
{
	{XtNlabel, XtCLabel, XtRString, sizeof(String),
	 offset(label), XtRString, NULL},
	{XtNvertSpace, XtCVertSpace, XtRInt, sizeof(int),
	 offset(vert_space), XtRImmediate, (XtPointer) 25},
	{XtNleftBitmap, XtCLeftBitmap, XtRBitmap, sizeof(Pixmap),
	 offset(left_bitmap), XtRImmediate, (XtPointer) None},
	{XtNjustify, XtCJustify, XtRJustify, sizeof(XtJustify),
	 offset(justify), XtRImmediate, (XtPointer) XtJustifyLeft},
	{XtNrightBitmap, XtCRightBitmap, XtRBitmap, sizeof(Pixmap),
	 offset(right_bitmap), XtRImmediate, (XtPointer) None},
	{XtNleftMargin, XtCHorizontalMargins, XtRDimension, sizeof(Dimension),
	 offset(left_margin), XtRImmediate, (XtPointer) 4},
  {XtNrightMargin, XtCHorizontalMargins, XtRDimension, sizeof(Dimension),
   offset(right_margin), XtRImmediate, (XtPointer) 4},
	{XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	 offset(foreground), XtRString, XtDefaultForeground},
	{XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	 offset(font), XtRString, XtDefaultFont},
	{XtNfontSet, XtCFontSet, XtRFontSet, sizeof(XFontSet),
	 offset(fontset), XtRString, XtDefaultFontSet},
};
#undef offset

/*
 * Semi Public function definitions. 
 */

static void Redisplay(), Destroy(), Initialize(), FlipColors();
static void ClassInitialize();
static Boolean SetValues();
static XtGeometryResult QueryGeometry();

/* 
 * Private Function Definitions.
 */

static void GetBitmapInfo(Widget w, Boolean is_left);
static void GetDefaultSize(), DrawBitmaps();
static void CreateGCs(), DestroyGCs();

#define superclass (&smeClassRec)
SmeCMPClassRec smeCMPClassRec =
{
	{
	/* superclass         */ (WidgetClass) superclass,
	/* class_name         */ "SmeCMP",
	/* size               */ sizeof(SmeCMPRec),
	/* class_initializer  */ ClassInitialize,
	/* class_part_initialize */ NULL,
	/* Class init'ed      */ FALSE,
	/* initialize         */ Initialize,
	/* initialize_hook    */ NULL,
	/* realize            */ NULL,
	/* actions            */ NULL,
	/* num_actions        */ ZERO,
	/* resources          */ resources,
	/* resource_count     */ XtNumber(resources),
	/* xrm_class          */ NULLQUARK,
	/* compress_motion    */ FALSE,
	/* compress_exposure  */ FALSE,
	/* compress_enterleave */ FALSE,
	/* visible_interest   */ FALSE,
	/* destroy            */ Destroy,
	/* resize             */ NULL,
	/* expose             */ Redisplay,
	/* set_values         */ SetValues,
	/* set_values_hook    */ NULL,
	/* set_values_almost  */ XtInheritSetValuesAlmost,
	/* get_values_hook    */ NULL,
	/* accept_focus       */ NULL,
	/* intrinsics version */ XtVersion,
	/* callback offsets   */ NULL,
	/* tm_table       */ NULL,
	/* query_geometry     */ QueryGeometry,
	/* display_accelerator */ NULL,
	/* extension      */ NULL
	},
	{
	/* SimpleMenuClass Fields */
	/* highlight          */ FlipColors,
	/* unhighlight        */ FlipColors,
	/* notify             */ XtInheritNotify,
	/* extension      */ NULL
	},
	{
	/* CMPClass Fields */
	/* extension      */ NULL
	}
};

WidgetClass smeCMPObjectClass = (WidgetClass) & smeCMPClassRec;

/************************************************************
 *
 * Semi-Public Functions.
 *
 ************************************************************/

/***************************************************************************
 *	ClassInitialize: SmeCMP IuWFNg class_initialize \bh
 *	`
 *		static void ClassInitialize(void);
 *	
 *		Initializes the SmeCMPObject. 
 *	QƎ
 *		X Toolkit Intrinsics Reference Manual   XtProc
 *												  class_initialize \bh
 *		X Toolkit Intrinsics Programing Manual  class_initialize \bh
 ***************************************************************************/
static void ClassInitialize(void)
{
	XawInitializeWidgetSet();
	XtAddConverter(XtRString, XtRJustify, XmuCvtStringToJustify,
				   (XtConvertArgList) NULL, (Cardinal) 0);
}

/***************************************************************************
 *	Initialize: SmeCMP IuWFNg initialize \bh
 *	`
 *		static void Initialize(Widget request, Widget new,
 *									ArgList args, Cardinal num_args)
 *	
 *		Widget request;		the widget requested by the argument list.
 *		Widget new;			the new widget with both resource and non
 *							resource values.
 *		ArgList args;		gp
 *		Cardinal num_args;	gp
 *	
 *		vCx[gCX^Xϐ̏
 *		pubNCX^Xϐ̗L̃`FbN
 *		
 *		ݒe
 *			
 ***************************************************************************/
/* ARGSUSED */
static void Initialize(Widget request, Widget new, 
									ArgList args, Cardinal *num_args)
{
	SmeCMPObject entry = (SmeCMPObject) new;

	if (entry->sme_cmp.label == NULL)
		entry->sme_cmp.label = XtName(new);
	else
		entry->sme_cmp.label = XtNewString(entry->sme_cmp.label);

	GetDefaultSize(new, &(entry->rectangle.width), &(entry->rectangle.height));
	CreateGCs(new);

	entry->sme_cmp.left_bitmap_width = entry->sme_cmp.left_bitmap_height = 0;
	entry->sme_cmp.right_bitmap_width = entry->sme_cmp.right_bitmap_height = 0;

	GetBitmapInfo(new, TRUE);	/* Left Bitmap Info */
	GetBitmapInfo(new, FALSE);	/* Right Bitmap Info */
}

/***************************************************************************
 *	Destroy: SmeCMP IuWFNg Destroy \bh
 *	`
 *		static void Destroy(Widget w);
 *	
 *		Widget w;	the simple menu widget.
 *	
 *		Called at destroy time, cleans up.
 ***************************************************************************/
static void Destroy(w)
Widget w;
{
	SmeCMPObject entry = (SmeCMPObject) w;

	DestroyGCs(w);
	if (entry->sme_cmp.label != XtName(w))
		XtFree(entry->sme_cmp.label);
}

/***************************************************************************
 *	Redisplay: SmeCMP IuWFNg expose \bh
 *	`
 *		static void Redisplay(Widget w, XEvent *event, Region region);
 *	
 *		Widget w;		SimpleMenu EBWFbg
 *		XEvent *event;	ĕ\Kvȋ`̃GNX|[YCxgw肵܂
 *		Region region;	ÃGNX|[XCxĝׂĂ킹`
 *						w肵܂
 *	
 *		@Redisplays the contents of the widget.
 ***************************************************************************/
/* ARGSUSED */
static void Redisplay(Widget w, XEvent *event, Region region)
{
	GC gc;
	SmeCMPObject entry = (SmeCMPObject) w;
	int font_ascent, font_descent, y_loc;

	int fontset_ascent, fontset_descent;
	XFontSetExtents *ext = XExtentsOfFontSet(entry->sme_cmp.fontset);

	entry->sme_cmp.set_values_area_cleared = FALSE;
	if (entry->sme.international == True) {
		fontset_ascent = abs(ext->max_ink_extent.y);
		fontset_descent = ext->max_ink_extent.height - fontset_ascent;
	} else {					/*else, compute size from font like R5 */
		font_ascent = entry->sme_cmp.font->max_bounds.ascent;
		font_descent = entry->sme_cmp.font->max_bounds.descent;
	}
	y_loc = entry->rectangle.y;


	if (XtIsSensitive(w) && XtIsSensitive(XtParent(w))) {
		if (w == XawSimpleMenuGetActiveEntry(XtParent(w))) {
			XFillRectangle(XtDisplayOfObject(w), XtWindowOfObject(w),
						   entry->sme_cmp.norm_gc, 0, y_loc,
						   (unsigned int) entry->rectangle.width,
						   (unsigned int) entry->rectangle.height);
			gc = entry->sme_cmp.rev_gc;
		} else
			gc = entry->sme_cmp.norm_gc;
	} else
		gc = entry->sme_cmp.norm_gray_gc;

	if (entry->sme_cmp.label != NULL) {
		int x_loc = entry->sme_cmp.left_margin;
		int len = strlen(entry->sme_cmp.label);
		char *label = entry->sme_cmp.label;

		switch (entry->sme_cmp.justify) {
			int width, t_width;

		case XtJustifyCenter:
			if (entry->sme.international == True) {
				t_width = XmbTextEscapement(entry->sme_cmp.fontset, label, len);
				width = entry->rectangle.width - (entry->sme_cmp.left_margin +
											entry->sme_cmp.right_margin);
			} else {
				t_width = XTextWidth(entry->sme_cmp.font, label, len);
				width = entry->rectangle.width - (entry->sme_cmp.left_margin +
											entry->sme_cmp.right_margin);
			}
			x_loc += (width - t_width) / 2;
			break;
		case XtJustifyRight:
			if (entry->sme.international == True) {
				t_width = XmbTextEscapement(entry->sme_cmp.fontset, label, len);
				x_loc = entry->rectangle.width - (entry->sme_cmp.right_margin
												  + t_width);
			} else {
				t_width = XTextWidth(entry->sme_cmp.font, label, len);
				x_loc = entry->rectangle.width - (entry->sme_cmp.right_margin
												  + t_width);
			}
			break;
		case XtJustifyLeft:
		default:
			break;
		}


		/* this will center the text in the gadget top-to-bottom */

		if (entry->sme.international == True) {
			y_loc += ((int) entry->rectangle.height -
				(fontset_ascent + fontset_descent)) / 2 + fontset_ascent;

			XmbDrawString(XtDisplayOfObject(w), XtWindowOfObject(w),
				   entry->sme_cmp.fontset, gc, x_loc, y_loc, label, len);
		} else {
			y_loc += ((int) entry->rectangle.height -
					  (font_ascent + font_descent)) / 2 + font_ascent;

			XDrawString(XtDisplayOfObject(w), XtWindowOfObject(w), gc,
						x_loc, y_loc, label, len);
		}
	}
	DrawBitmaps(w, gc);
}


/***************************************************************************
 *	SetValues: SmeCMP IuWFNg set_values \bh
 *	`
 *		static Boolean SetValues(Widget current, Widget request, 
 *					Widget new, ArgList args, Cardinal *num_args);
 *	
 *		Widget current;		XtSetValues() ĂяoȌԂ̃EBWFbg
 *							Rs[
 *		Widget request;		XtSetValues() ĂяoĒlύXꂽ
 *							̃NX set_values() ֐ĂяoĂ
 *							Ԃ̃EBWFbg̃Rs[
 *		Widget new;			łɐݒ肳Ă set_values() ĂяoŏC
 *							ꂽEBWFbg
 *		ArgList args;		XtSetValues() ĂяoɓnꂽXg
 *		Cardinal *num_args;	Xg̈̐
 *	Ԃl
 *		FalseiUj: 	ECWFbgĕ`悷
 *		True i^j: 	ECWFbgĕ`悵Ȃ
 *	
 *		@ȉ̃\[XɂāAĕ`̕Kv邩𔻒肷B
 *		
 *		\[XF	EBWFbg
 *		\[XF	EBWFbg
 *		\[XF	
 *		
 *		Relayout the menu when one of the resources is changed.
 *	QƎ
 *		See. X Toolkit Intrinsics Reference Manual		XtSetValuesFunc
 *														set_values \bh
 *			 X Toolkit Intrinsics Programing Manual		set_values \bh
 ***************************************************************************/
/* ARGSUSED */
static Boolean SetValues(Widget current, Widget request, Widget new, 
										ArgList args, Cardinal *num_args)
{
	SmeCMPObject entry = (SmeCMPObject) new;
	SmeCMPObject old_entry = (SmeCMPObject) current;
	Boolean ret_val = FALSE;

	if (old_entry->sme_cmp.label != entry->sme_cmp.label) {
		if (old_entry->sme_cmp.label != XtName(new))
			XtFree((char *) old_entry->sme_cmp.label);

		if (entry->sme_cmp.label != XtName(new))
			entry->sme_cmp.label = XtNewString(entry->sme_cmp.label);

		ret_val = True;
	}
	if (entry->rectangle.sensitive != old_entry->rectangle.sensitive)
		ret_val = TRUE;

	if (entry->sme_cmp.left_bitmap != old_entry->sme_cmp.left_bitmap) {
		GetBitmapInfo(new, TRUE);
		ret_val = TRUE;
	}
	if (entry->sme_cmp.right_bitmap != old_entry->sme_cmp.right_bitmap) {
		GetBitmapInfo(new, FALSE);
		ret_val = TRUE;
	}
	if (((old_entry->sme_cmp.font != entry->sme_cmp.font) &&
		 (old_entry->sme.international == False)) ||
		(old_entry->sme_cmp.foreground != entry->sme_cmp.foreground)) {
		DestroyGCs(current);
		CreateGCs(new);
		ret_val = TRUE;
	}
	if ((old_entry->sme_cmp.fontset != entry->sme_cmp.fontset) &&
		(old_entry->sme.international == True))
		/* DONT changes the GCs, because the fontset is not in them. */
		ret_val = TRUE;

	if (ret_val) {
		GetDefaultSize(new,
				  &(entry->rectangle.width), &(entry->rectangle.height));
		entry->sme_cmp.set_values_area_cleared = TRUE;
	}
	return (ret_val);
}

/***************************************************************************
 *	QueryGeometry: SmeCMP IuWFNg query_geometry \bh
 *	`
 *		static XtGeometryResult QueryGeometry(Widget w, 
 *				XtWidgetGeometry *intended, XtWidgetGeometry *preferred);
 *	
 *		íj
 *		Widget w;						the menu entry object.
 *		XtWidgetGeometry *intended;		esƂĂWIg̕ύX
 *		iój
 *		XtWidgetGeometry *return_val;	EBWFbgIuWFNg̓K؂
 *										WIg
 *	Ԃl
 *		A Geometry Result
 *	
 *		Returns the preferred geometry for this widget.
 *	QƎ
 *		See the Intrinsics manual for details on what this function is for.
 *
 *		I just return the height and width of the label plus the margins.
 ***************************************************************************/
static XtGeometryResult QueryGeometry(Widget w, 
				XtWidgetGeometry *intended, XtWidgetGeometry *return_val)
{
	SmeCMPObject entry = (SmeCMPObject) w;
	Dimension width, height;
	XtGeometryResult ret_val = XtGeometryYes;
	XtGeometryMask mode = intended->request_mode;

	GetDefaultSize(w, &width, &height);

	if (((mode & CWWidth) && (intended->width != width)) ||
		!(mode & CWWidth)) {
		return_val->request_mode |= CWWidth;
		return_val->width = width;
		ret_val = XtGeometryAlmost;
	}
	if (((mode & CWHeight) && (intended->height != height)) ||
		!(mode & CWHeight)) {
		return_val->request_mode |= CWHeight;
		return_val->height = height;
		ret_val = XtGeometryAlmost;
	}
	if (ret_val == XtGeometryAlmost) {
		mode = return_val->request_mode;

		if (((mode & CWWidth) && (width == entry->rectangle.width)) &&
			((mode & CWHeight) && (height == entry->rectangle.height)))
			return (XtGeometryNo);
	}
	return (ret_val);
}

/***************************************************************************
 *	FlipColors: highlight()/unhighlight() ANV
 *	`
 *		static void FlipColors(Widget w);
 *	
 *		Widget w;		the cmp menu entry widget.
 *	
 *		@Redisplays the contents of the widget.
 ***************************************************************************/
static void FlipColors(Widget w)
{
	SmeCMPObject entry = (SmeCMPObject) w;
	SmeCMPObjectClass oclass = (SmeCMPObjectClass) XtClass(w);

	if (entry->sme_cmp.set_values_area_cleared) {
		return;
	}
	XDrawRectangle(XtDisplayOfObject(w), XtWindowOfObject(w),
				   entry->sme_cmp.invert_gc,
				   0, (int) entry->rectangle.y,
				   (unsigned int) entry->rectangle.width-1,
				   (unsigned int) entry->rectangle.height-1);
}

/************************************************************
 *
 *	vCx[g֐Q
 *
 ************************************************************/

/***************************************************************************
 *	GetDefaultSize: 
 *	`
 *		static void GetDefaultSize(Widget w, 
 *						Dimension *width, Dimension *height);
 *	
 *		Widget w;			the menu entry widget.
 *		Dimension *width;	default sizes (RETURNED).
 *		Dimension *height;	default sizes (RETURNED).
 *	
 *		@Calculates the Default (preferred) size of this menu entry.
 ***************************************************************************/
static void GetDefaultSize(Widget w, Dimension *width, Dimension *height)
{
	SmeCMPObject entry = (SmeCMPObject) w;

	if (entry->sme.international == True) {
		XFontSetExtents *ext = XExtentsOfFontSet(entry->sme_cmp.fontset);
		if (entry->sme_cmp.label == NULL)
			*width = 0;
		else
			*width = XmbTextEscapement(entry->sme_cmp.fontset, 
						entry->sme_cmp.label, strlen(entry->sme_cmp.label));
		*width += entry->sme_cmp.left_margin + entry->sme_cmp.right_margin;
		*height = ext->max_ink_extent.height;
		*height = ((int) *height * (ONE_HUNDRED +
							   entry->sme_cmp.vert_space)) / ONE_HUNDRED;
	} else {
		if (entry->sme_cmp.label == NULL)
			*width = 0;
		else
			*width = XTextWidth(entry->sme_cmp.font, entry->sme_cmp.label,
								strlen(entry->sme_cmp.label));

		*width += entry->sme_cmp.left_margin + entry->sme_cmp.right_margin;

		*height = (entry->sme_cmp.font->max_bounds.ascent +
				   entry->sme_cmp.font->max_bounds.descent);

		*height = ((int) *height * (ONE_HUNDRED +
							   entry->sme_cmp.vert_space)) / ONE_HUNDRED;
	}
}

/***************************************************************************
 *	DrawBitmaps: 
 *	`
 *		static void DrawBitmaps(Widget w, GC gc);
 *	
 *		Widget w;		the simple menu widget.
 *		GC gc;			graphics context to use for drawing.
 *	
 *		@Draws left and right bitmaps.
 ***************************************************************************/
static void DrawBitmaps(Widget w, GC gc)
{
	int x_loc, y_loc;
	SmeCMPObject entry = (SmeCMPObject) w;

	if ((entry->sme_cmp.left_bitmap == None) &&
		(entry->sme_cmp.right_bitmap == None))
		return;

/*
 * Draw Left Bitmap.
 */

	if (entry->sme_cmp.left_bitmap != None) {
		x_loc = (int) (entry->sme_cmp.left_margin -
					   entry->sme_cmp.left_bitmap_width) / 2;

		y_loc = entry->rectangle.y +
			(int) (entry->rectangle.height -
				   entry->sme_cmp.left_bitmap_height) / 2;

		XCopyPlane(XtDisplayOfObject(w), entry->sme_cmp.left_bitmap,
				   XtWindowOfObject(w), gc, 0, 0,
				   entry->sme_cmp.left_bitmap_width,
				   entry->sme_cmp.left_bitmap_height, x_loc, y_loc, 1);
	}
/*
 * Draw Right Bitmap.
 */


	if (entry->sme_cmp.right_bitmap != None) {
		x_loc = entry->rectangle.width -
			(int) (entry->sme_cmp.right_margin +
				   entry->sme_cmp.right_bitmap_width) / 2;

		y_loc = entry->rectangle.y +
			(int) (entry->rectangle.height -
				   entry->sme_cmp.right_bitmap_height) / 2;

		XCopyPlane(XtDisplayOfObject(w), entry->sme_cmp.right_bitmap,
				   XtWindowOfObject(w), gc, 0, 0,
				   entry->sme_cmp.right_bitmap_width,
				   entry->sme_cmp.right_bitmap_height, x_loc, y_loc, 1);
	}
}

/***************************************************************************
 *	GetBitmapInfo: 
 *	`
 *		static void GetBitmapInfo(Widget w, Boolean is_left);
 *	
 *		Widget w;			the cmp menu entry widget.
 *		Boolean is_left;	TRUE if we are testing left bitmap,
 *							FALSE if we are testing the right bitmap.
 *	
 *		@Gets the bitmap information from either of the bitmaps.
 ***************************************************************************/
static void GetBitmapInfo(Widget w, Boolean is_left)
{
	SmeCMPObject entry = (SmeCMPObject) w;
	unsigned int depth, bw;
	Window root;
	int x, y;
	unsigned int width, height;
	char buf[BUFSIZ];

	if (is_left) {
		if (entry->sme_cmp.left_bitmap != None) {
			if (!XGetGeometry(XtDisplayOfObject(w),
							  entry->sme_cmp.left_bitmap, &root,
							  &x, &y, &width, &height, &bw, &depth)) {
				(void) sprintf(buf, "Xaw SmeCMP Object: %s %s \"%s\".", 
					"Could not",
				  	"get Left Bitmap geometry information for menu entry ",
							   XtName(w));
				XtAppError(XtWidgetToApplicationContext(w), buf);
			}
			if (depth != 1) {
				(void) sprintf(buf, "Xaw SmeCMP Object: %s \"%s\"%s.",
							   "Left Bitmap of entry ",
							   XtName(w), " is not one bit deep.");
				XtAppError(XtWidgetToApplicationContext(w), buf);
			}
			entry->sme_cmp.left_bitmap_width = (Dimension) width;
			entry->sme_cmp.left_bitmap_height = (Dimension) height;
		}
	} else if (entry->sme_cmp.right_bitmap != None) {
		if (!XGetGeometry(XtDisplayOfObject(w),
						  entry->sme_cmp.right_bitmap, &root,
						  &x, &y, &width, &height, &bw, &depth)) {
			(void) sprintf(buf, "Xaw SmeCMP Object: %s %s \"%s\".", 
					"Could not",
					"get Right Bitmap geometry information for menu entry ",
						   XtName(w));
			XtAppError(XtWidgetToApplicationContext(w), buf);
		}
		if (depth != 1) {
			(void) sprintf(buf, "Xaw SmeCMP Object: %s \"%s\"%s.",
						   "Right Bitmap of entry ", XtName(w),
						   " is not one bit deep.");
			XtAppError(XtWidgetToApplicationContext(w), buf);
		}
		entry->sme_cmp.right_bitmap_width = (Dimension) width;
		entry->sme_cmp.right_bitmap_height = (Dimension) height;
	}
}

/***************************************************************************
 *	CreateGCs: 
 *	`
 *		static void CreateGCs(Widget w);
 *	
 *		Widget w;		the simple menu widget.
 *	
 *		@Creates all gc's for the simple menu widget.
 ***************************************************************************/
static void CreateGCs(Widget w)
{
	SmeCMPObject entry = (SmeCMPObject) w;
	XGCValues values;
	XtGCMask mask, mask_i18n;

	values.foreground = XtParent(w)->core.background_pixel;
	values.background = entry->sme_cmp.foreground;
	values.font = entry->sme_cmp.font->fid;
	values.graphics_exposures = FALSE;
	mask = GCForeground | GCBackground | GCGraphicsExposures | GCFont;
	mask_i18n = GCForeground | GCBackground | GCGraphicsExposures;
	if (entry->sme.international == True)
		entry->sme_cmp.rev_gc = XtAllocateGC(w, 0, mask_i18n, &values, GCFont, 0);
	else
		entry->sme_cmp.rev_gc = XtGetGC(w, mask, &values);

	values.foreground = entry->sme_cmp.foreground;
	values.background = XtParent(w)->core.background_pixel;
	if (entry->sme.international == True)
		entry->sme_cmp.norm_gc = XtAllocateGC(w, 0, mask_i18n, &values, GCFont, 0);
	else
		entry->sme_cmp.norm_gc = XtGetGC(w, mask, &values);

	values.fill_style = FillTiled;
	values.tile = XmuCreateStippledPixmap(XtScreenOfObject(w),
										  entry->sme_cmp.foreground,
									  XtParent(w)->core.background_pixel,
										  XtParent(w)->core.depth);
	values.graphics_exposures = FALSE;
	mask |= GCTile | GCFillStyle;
	if (entry->sme.international == True)
		entry->sme_cmp.norm_gray_gc = XtAllocateGC(w, 0, mask_i18n, &values, GCFont, 0);
	else
		entry->sme_cmp.norm_gray_gc = XtGetGC(w, mask, &values);

	values.foreground ^= values.background;
	values.background = 0;
	values.function = GXxor;
	mask = GCForeground | GCBackground | GCGraphicsExposures | GCFunction;
	entry->sme_cmp.invert_gc = XtGetGC(w, mask, &values);
}

/***************************************************************************
 *	DestroyGCs: 
 *	`
 *		static void DestroyGCs(Widget w);
 *	
 *		Widget w;		the simple menu widget.
 *	
 *		@Removes all gc's for the simple menu widget.
 ***************************************************************************/
static void DestroyGCs(Widget w)
{
	SmeCMPObject entry = (SmeCMPObject) w;

	XtReleaseGC(w, entry->sme_cmp.norm_gc);
	XtReleaseGC(w, entry->sme_cmp.norm_gray_gc);
	XtReleaseGC(w, entry->sme_cmp.rev_gc);
	XtReleaseGC(w, entry->sme_cmp.invert_gc);
}

#ifdef apollo

/*
 * The apollo compiler that we have optomizes out my code for
 * FlipColors() since it is static. and no one executes it in this
 * file.  I am setting the function pointer into the class structure so
 * that it can be called by my parent who will tell me to when to
 * highlight and unhighlight.
 */

void _XawSmeCMPApolloHack()
{
	FlipColors();
}

#endif							/* apollo */
